/**
 * Copyright © schukai GmbH and all contributing authors, {{copyRightYear}}. All rights reserved.
 * Node module: @schukai/monster
 *
 * This source code is licensed under the GNU Affero General Public License version 3 (AGPLv3).
 * The full text of the license can be found at: https://www.gnu.org/licenses/agpl-3.0.en.html
 *
 * For those who do not wish to adhere to the AGPLv3, a commercial license is available.
 * Acquiring a commercial license allows you to use this software without complying with the AGPLv3 terms.
 * For more information about purchasing a commercial license, please contact schukai GmbH.
 *
 * SPDX-License-Identifier: AGPL-3.0
 */

import { AbstractOperator } from "./abstractoperator.mjs";
import { instanceSymbol } from "../constants.mjs";
export { OrOperator };

/**
 * Constraints are used to define conditions that must be met by the value of a variable.
 *
 * The uniform API of the constraints allows chains to be formed.
 *
 * The OrOperator is used to link several constraints. The constraint is fulfilled if one of the constraints is fulfilled.
 *
 * @example /examples/libraries/constraints/oroperator/ A simple example
 *
 * @license AGPLv3
 * @since 1.3.0
 * @copyright schukai GmbH
 * @summary A or operator
 */
class OrOperator extends AbstractOperator {
	/**
	 * this method returns a promise containing the result of the check.
	 *
	 * @param {*} value
	 * @return {Promise}
	 */
	isValid(value) {
		var self = this;

		return new Promise(function (resolve, reject) {
			let a;
			let b;

			self.operantA
				.isValid(value)
				.then(function () {
					resolve();
				})
				.catch(function () {
					a = false;
					/** b has already been evaluated and was not true */
					if (b === false) {
						reject();
					}
				});

			self.operantB
				.isValid(value)
				.then(function () {
					resolve();
				})
				.catch(function () {
					b = false;
					/** b has already been evaluated and was not true */
					if (a === false) {
						reject();
					}
				});
		});
	}

	/**
	 * This method is called by the `instanceof` operator.
	 * @return {symbol}
	 * @since 2.1.0
	 */
	static get [instanceSymbol]() {
		return Symbol.for("@schukai/monster/constraints/or-operator");
	}
}
